MODULE cubefile
    IMPLICIT NONE
    PRIVATE
    REAL :: origin(3),trust_radius=3.0,diagonal_length
    REAL,PARAMETER :: Temperature=298.0
    INTEGER,PARAMETER :: Npositions=1000,dummy=Npositions
    REAL :: outer(3),middle(3),inner(3),gridsize(3)
    REAL,PARAMETER :: force_constant_default=1000.0
    INTEGER :: positions(Npositions+1,3)=1,stride=10
    REAL,DIMENSION(:,:,:),ALLOCATABLE  :: cube !cube dimensions are fast-middle-slow, or equivalently z-y-x
    INTEGER,DIMENSION(:,:,:),ALLOCATABLE  :: working_cube 
    INTEGER,DIMENSION(:),ALLOCATABLE  :: NEB_string
    REAL,DIMENSION(:,:),ALLOCATABLE :: positions_real
    INTEGER :: NAtoms,Nouter,Nmiddle,Ninner
    PUBLIC :: initialise_cubefile,finalise_cubefile,set_vector,fill_maxima,get_vector,read_vectors,print_NEBSTRING_coordinates
    PUBLIC :: maximise_vector,interpolate_vectors,NEB_segment,full_NEB,invert_cubevalues,print_values_global,print_coordinates
	PUBLIC :: Temperature
    CONTAINS

        SUBROUTINE initialise_cubefile(filename)
        IMPLICIT NONE
        CHARACTER(LEN=*),INTENT(IN) :: filename
        LOGICAL :: file_exists,connected
        INTEGER :: allocstatus,ios
            CALL srand(86456)
            INQUIRE(FILE=TRIM(filename),EXIST=file_exists)
            IF (file_exists) THEN
                WRITE(*,'(" opening cube file: ",A)') "'"//TRIM(filename)//"'"
                INQUIRE(UNIT=3,OPENED=connected)
                IF (connected) CALL error_handler("unit 3 already connected")
                OPEN(UNIT=3,FILE=TRIM(filename),ACTION='READ',IOSTAT=ios)
                IF (ios/=0) CALL error_handler("can't open cube file")
                CALL read_header()
                ALLOCATE(cube(Ninner,Nmiddle,Nouter),STAT=allocstatus)
                IF (allocstatus/=0) CALL error_handler("cannot allocate cube memory")
                ALLOCATE(working_cube(Ninner,Nmiddle,Nouter),STAT=allocstatus)
                IF (allocstatus/=0) CALL error_handler("cannot allocate working_cube memory")
                CALL read_body()
                positions(:,:)=0
            ELSE
                CALL error_handler("file doesn't exist")
            ENDIF
        END SUBROUTINE initialise_cubefile

        SUBROUTINE finalise_cubefile()
        IMPLICIT NONE
        LOGICAL :: connected
        INTEGER :: deallocstatus
            INQUIRE(UNIT=3,OPENED=connected)
            IF (connected) CLOSE(UNIT=3)
            IF (ALLOCATED(cube)) THEN
                DEALLOCATE(cube,STAT=deallocstatus)
                IF (deallocstatus/=0) CALL error_handler("cannot deallocate cube memory")
            ENDIF
            IF (ALLOCATED(working_cube)) THEN
                DEALLOCATE(working_cube,STAT=deallocstatus)
                IF (deallocstatus/=0) CALL error_handler("cannot deallocate working_cube memory")
            ENDIF
            IF (ALLOCATED(positions_real)) THEN
                DEALLOCATE(positions_real,STAT=deallocstatus)
                IF (deallocstatus/=0) CALL error_handler("cannot deallocate positions_real memory")
            ENDIF
        END SUBROUTINE finalise_cubefile

        SUBROUTINE read_header()
        IMPLICIT NONE
        INTEGER :: ios,n
            REWIND 3
            READ(3,IOSTAT=ios,FMT=*)
            IF (ios/=0) CALL error_handler("can't read first line")
            READ(3,IOSTAT=ios,FMT=*)
            IF (ios/=0) CALL error_handler("can't read second line")
            READ(3,IOSTAT=ios,FMT=*) NAtoms,origin
            IF (ios/=0) CALL error_handler("can't read NAtoms / origin")
            READ(3,IOSTAT=ios,FMT=*) Nouter,outer ! is X in general.
            IF (ios/=0) CALL error_handler("can't read Nouter / outer")
            READ(3,IOSTAT=ios,FMT=*) Nmiddle,middle ! is Y
            IF (ios/=0) CALL error_handler("can't read Nmiddle / middle")
            READ(3,IOSTAT=ios,FMT=*) Ninner,inner ! is Z
            IF (ios/=0) CALL error_handler("can't read Ninner / inner")
            DO n=1,NAtoms,1
                READ(3,IOSTAT=ios,FMT=*)
                IF (ios/=0) CALL error_handler("can't read atom lines")
            ENDDO
            WRITE(*,*) "Switching from Bohr to Angströms"
            origin=origin*0.529177211
            outer=outer*0.529177211
            middle=middle*0.529177211
            inner=inner*0.529177211
            gridsize(1)=SQRT(SUM(outer(:)**2))
            gridsize(2)=SQRT(SUM(middle(:)**2))
            gridsize(3)=SQRT(SUM(inner(:)**2))
            diagonal_length=SQRT(SUM(gridsize(:)**2))*1.001
            WRITE(*,*) "Notice: reading molecules not yet implemented."
        END SUBROUTINE read_header

        SUBROUTINE read_body()
        IMPLICIT NONE
        INTEGER :: ios
            WRITE(*,'(" Reading ",I0,"x",I0,"x",I0," records.")') Ninner,Nmiddle,Nouter
            READ(3,IOSTAT=ios,FMT=*) cube(:,:,:)
            IF (ios/=0) CALL error_handler("cannot read record...")
            WRITE(*,'(" first, second, and last record: ")')
            WRITE(*,*) cube(1,1,1),cube(2,1,1),cube(Ninner,Nmiddle,Nouter)
            WRITE(*,'(" minimum and maximum: ")')
            WRITE(*,*) MINVAL(cube),MAXVAL(cube)
            WRITE(*,'(" maximum position: ")')
            WRITE(*,*) MAXLOC(cube)
            WRITE(*,*) FLOAT(MAXLOC(cube))*outer(1)+origin(1)
        END SUBROUTINE read_body

        REAL FUNCTION givevalue_positions(positions_in)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: positions_in(3)
            givevalue_positions=cube(positions_in(3),positions_in(2),positions_in(1))
        END FUNCTION givevalue_positions

        REAL FUNCTION givevalue_positions_averaged(positions_in)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: positions_in(3)
        INTEGER :: a,b,c
            givevalue_positions_averaged=0.0
            DO a=-1,1,1
                DO b=-1,1,1
                    DO c=-1,1,1
                        givevalue_positions_averaged=givevalue_positions_averaged+&
                        &cube(positions_in(3)+a,positions_in(2)+b,positions_in(1)+c)
                    ENDDO
                ENDDO
            ENDDO
            givevalue_positions_averaged=givevalue_positions_averaged/27.0
        END FUNCTION givevalue_positions_averaged

        REAL FUNCTION givevalue_index(position_index)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: position_index
            givevalue_index=cube(positions(position_index,3),positions(position_index,2),positions(position_index,1))
        END FUNCTION givevalue_index

        REAL FUNCTION givevalue_index_averaged(position_index)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: position_index
        INTEGER :: a,b,c
            givevalue_index_averaged=0.0
            DO a=-1,1,1
                DO b=-1,1,1
                    DO c=-1,1,1
                        IF (.NOT.(valid_position_d3(positions(position_index,:)))) THEN
                            WRITE(*,*) "exceeded valid range ",get_vector(position_index),positions(position_index,:),position_index
                            givevalue_index_averaged=0.0
                        ELSE
                            givevalue_index_averaged=givevalue_index_averaged+&
                            &cube(positions(position_index,3)+a,positions(position_index,2)+b,positions(position_index,1)+c)
                        ENDIF
                    ENDDO
                ENDDO
            ENDDO
            givevalue_index_averaged=givevalue_index_averaged/27.0
        END FUNCTION givevalue_index_averaged

        RECURSIVE SUBROUTINE maximise_vector(position_index,shifting,trust_radius_in)
        IMPLICIT NONE
        REAL,OPTIONAL,INTENT(IN) :: trust_radius_in
        LOGICAL,INTENT(IN) :: shifting
        INTEGER,INTENT(IN) :: position_index
        REAL :: initial_vector(3),best_value
            IF (PRESENT(trust_radius_in)) trust_radius=trust_radius_in
            initial_vector=get_vector(position_index)
            best_value=givevalue_positions(positions(position_index,:))
            WRITE(*,*) "maximising..."
            WRITE(*,*) "initial vector is ",initial_vector
            working_cube(:,:,:)=0
            CALL REC_maximise_vector(positions(position_index,:))
            WRITE(*,*) "final vector is ",get_vector(position_index)
            IF (shifting) THEN
                !update initial vector, reset working_cube
                IF (SUM((initial_vector-get_vector(position_index))**2)<0.001) THEN
                    WRITE(*,*) "Converged."
                ELSE
                    WRITE(*,*) "Resetting..."
                    CALL maximise_vector(position_index,shifting,trust_radius_in)
                ENDIF
            ENDIF

            CONTAINS

                RECURSIVE SUBROUTINE REC_maximise_vector(current_position)
                IMPLICIT NONE
                INTEGER,INTENT(IN) :: current_position(3)
                REAL :: distance,clipboard,current_vector(3)
                INTEGER :: a,b,c,shift(3)
                    !stop recursion if:
                    ! - the box boundaries are exceeded.
                    ! - the trust radius is left
                    ! - the tile has already been visited
                    IF (.NOT.(valid_position_d3(current_position))) RETURN
                    current_vector=get_vector_3d(current_position)
                    distance=SUM(SQRT((initial_vector(:)-current_vector(:))**2))
                    IF (distance>trust_radius) RETURN
                    IF (working_cube(current_position(3),current_position(2),current_position(1))==0) THEN
                        working_cube(current_position(3),current_position(2),current_position(1))=1
                        clipboard=cube(current_position(3),current_position(2),current_position(1))
                        IF (clipboard>best_value) THEN
                            WRITE(*,*) "new best: ",clipboard
                            best_value=clipboard
                            positions(position_index,:)=current_position
                        ENDIF
                        DO a=-1,1,1
                            shift(3)=a
                            DO b=-1,1,1
                                shift(2)=b
                                DO c=-1,1,1
                                    shift(1)=c
                                    CALL REC_maximise_vector(current_position+shift)
                                ENDDO
                            ENDDO
                        ENDDO
                    ENDIF
                END SUBROUTINE REC_maximise_vector

        END SUBROUTINE maximise_vector

        LOGICAL FUNCTION valid_position_d3(positions_in)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: positions_in(3)
            valid_position_d3=.TRUE.
            IF ((positions_in(1)<1).OR.(positions_in(1)>Nouter)) valid_position_d3=.FALSE.
            IF ((positions_in(2)<1).OR.(positions_in(2)>Nmiddle)) valid_position_d3=.FALSE.
            IF ((positions_in(2)<1).OR.(positions_in(2)>Ninner)) valid_position_d3=.FALSE.
        END FUNCTION valid_position_d3

        SUBROUTINE set_vector(input_vector,position_index)
        IMPLICIT NONE
        REAL,INTENT(IN) :: input_vector(3)
        INTEGER,INTENT(IN) :: position_index
            !WRITE(*,*) "Notice: assuming pure x,y,z vectors. No projection yet."
            IF (position_index>Npositions) CALL error_handler("exceeded allowed number of positions.")
            positions(position_index,:)=INT((input_vector(:)-origin(:))/gridsize(:))
            IF (positions(position_index,1)<1) CALL error_handler("exceeded box lowb in 'outer' (commonly 'x').")
            IF (positions(position_index,2)<1) CALL error_handler("exceeded box lowb in 'middle' (commonly 'y').")
            IF (positions(position_index,3)<1) CALL error_handler("exceeded box lowb in 'inner' (commonly 'z').")
            IF (positions(position_index,1)>Nouter) CALL error_handler("exceeded box upper in 'outer' (commonly 'x').")
            IF (positions(position_index,2)>Nmiddle) CALL error_handler("exceeded box upper in 'middle' (commonly 'y').")
            IF (positions(position_index,3)>Ninner) CALL error_handler("exceeded box upper in 'inner' (commonly 'z').")
            !WRITE(*,*) "Transformed Vector: ",input_vector(:)
            !WRITE(*,*) "to positions: ",positions(position_index,:)
            !WRITE(*,*) "cubefile: ",givevalue_positions(positions(position_index,:))
        END SUBROUTINE set_vector

        FUNCTION get_vector_3d(positions_in)
        IMPLICIT NONE
        REAL :: get_vector_3d(3)
        INTEGER,INTENT(IN) :: positions_in(3)
            get_vector_3d(:)=FLOAT(positions_in(:))*outer(1)+origin(1)
        END FUNCTION get_vector_3d

        FUNCTION get_vector(position_index)
        IMPLICIT NONE
        REAL :: get_vector(3)
        INTEGER,INTENT(IN) :: position_index
            get_vector(:)=FLOAT(positions(position_index,:))*outer(1)+origin(1)
        END FUNCTION get_vector

        SUBROUTINE fill_maxima(Niterations)
        IMPLICIT NONE
        INTEGER :: maximum_vector(3),iterations,ios
        REAL :: initial_vector(3)
        INTEGER,INTENT(IN) :: Niterations
        LOGICAL :: connected
            INQUIRE(UNIT=9,OPENED=connected)
            IF (connected) CALL error_handler("unit 9 already connected")
            OPEN(UNIT=9,FILE="BASINS.xyz",ACTION='WRITE',IOSTAT=ios)
            IF (ios/=0) CALL error_handler("can't open output file")
            REWIND 9
            WRITE(9,*) Niterations
            WRITE(9,*)
            DO iterations=1,Niterations,1
                maximum_vector=MAXLOC(cube)
                CALL switchxzint(maximum_vector)
                initial_vector(:)=get_vector_3d(maximum_vector)
                CALL REC_fill_maxima(maximum_vector)
                WRITE(9,*) "X ",initial_vector(:)
            ENDDO
            INQUIRE(UNIT=9,OPENED=connected)
            IF (connected) THEN
                CLOSE(UNIT=9)
            ELSE
                CALL error_handler("can't close output file")
            ENDIF

            CONTAINS

                RECURSIVE SUBROUTINE REC_fill_maxima(current_position)
                IMPLICIT NONE
                INTEGER,INTENT(IN) :: current_position(3)
                REAL :: distance,current_vector(3)
                INTEGER :: a,b,c,shift(3)
                    !stop recursion if:
                    ! - the box boundaries are exceeded.
                    ! - the trust radius is left
                    ! - the tile has already been visited
                    IF (.NOT.(valid_position_d3(current_position))) RETURN
                    current_vector=get_vector_3d(current_position)
                    distance=SUM(SQRT((initial_vector(:)-current_vector(:))**2))
                    IF (distance>trust_radius) RETURN
                    IF (cube(current_position(3),current_position(2),current_position(1))>-1) THEN
                        cube(current_position(3),current_position(2),current_position(1))=-1
                        DO a=-1,1,1
                            shift(3)=a
                            DO b=-1,1,1
                                shift(2)=b
                                DO c=-1,1,1
                                    shift(1)=c
                                    CALL REC_fill_maxima(current_position+shift)
                                ENDDO
                            ENDDO
                        ENDDO
                    ENDIF
                END SUBROUTINE REC_fill_maxima

        END SUBROUTINE fill_maxima

        SUBROUTINE error_handler(error)
        IMPLICIT NONE
        CHARACTER(LEN=*),INTENT(IN) :: error
            WRITE(*,'("### ERROR: ",A)') error
            STOP
        END SUBROUTINE error_handler

        SUBROUTINE invert_cubevalues()
        IMPLICIT NONE
            WRITE(*,*) "cube:=-cube"
            cube(:,:,:)=-cube(:,:,:)
        END SUBROUTINE invert_cubevalues

        SUBROUTINE switchxzint(inout_vector)
        IMPLICIT NONE
        INTEGER,INTENT(INOUT) :: inout_vector(3)
        INTEGER :: clipboard
            clipboard=inout_vector(1)
            inout_vector(1)=inout_vector(3)
            inout_vector(3)=clipboard
        END SUBROUTINE switchxzint

        SUBROUTINE interpolate_vectors(position_index_1,position_index_2)
        IMPLICIT NONE
        INTEGER,INTENT(IN) :: position_index_1,position_index_2
        INTEGER :: n
        REAL :: shift(3),initial_vector(3),final_vector(3)
            initial_vector=get_vector(position_index_1)
            final_vector=get_vector(position_index_2)
            shift=(final_vector-initial_vector)/FLOAT(position_index_2-position_index_1)
            DO n=1,position_index_2-position_index_1-1,1
                CALL set_vector(initial_vector+shift*FLOAT(n),position_index_1+n)
            ENDDO
        END SUBROUTINE interpolate_vectors

        SUBROUTINE NEB_segment(position_index_1,position_index_2,dynamic_force_constants,decreasing_stepsize,&
        &append_file,Niterations,decreasing_forceconstants,reuse_positions_in)
        IMPLICIT NONE
        LOGICAL,INTENT(IN) :: dynamic_force_constants,decreasing_stepsize,append_file,decreasing_forceconstants,reuse_positions_in
        INTEGER,INTENT(IN) :: position_index_1,position_index_2
        REAL :: force_constants(position_index_2-position_index_1-1)
        REAL :: tangents(position_index_2-position_index_1-1,3)
        REAL :: Fparallel(position_index_2-position_index_1-1,3)
        REAL :: Fperpendicular(position_index_2-position_index_1-1,3)
        REAL :: ForceRatio(position_index_2-position_index_1-1)
        REAL :: maxstep,max_displacement,average_force_constants
        INTEGER :: iterations,Niterations,deallocstatus
        LOGICAL :: connected,reuse_positions
            reuse_positions=reuse_positions_in
            WRITE(*,*) "Calling NEB"
            CALL initialise_NEB()
            DO iterations=0,Niterations,1
                CALL adjust_stepsize()
                CALL calculate_tangents()
                CALL calculate_forces()
                CALL displace()
                IF (MOD(iterations,stride)==0) CALL print_step()
            ENDDO
            IF (dynamic_force_constants) WRITE(*,*) "Average force constant: ",average_force_constants/Niterations
            INQUIRE(UNIT=9,OPENED=connected)
            IF (connected) CLOSE(UNIT=9)
            CALL print_values()
            IF ((ALLOCATED(positions_real)).AND.(.NOT.(reuse_positions))) THEN
                DEALLOCATE(positions_real,STAT=deallocstatus)
                IF (deallocstatus/=0) CALL error_handler("cannot deallocate positions_real memory")
            ENDIF

        CONTAINS

            SUBROUTINE adjust_stepsize()
            IMPLICIT NONE
                IF (decreasing_stepsize) maxstep=(diagonal_length/10.0)-(diagonal_length/(10.0*Niterations))*iterations
                IF (decreasing_forceconstants) force_constants(:)=(force_constant_default)-(0.0/Niterations)*iterations
            END SUBROUTINE adjust_stepsize

            SUBROUTINE initialise_NEB()
            IMPLICIT NONE
            INTEGER :: n,ios,allocstatus
            LOGICAL :: skip_position_initialisation
                average_force_constants=0.0
                INQUIRE(UNIT=9,OPENED=connected)
                IF (connected) CALL error_handler("unit 9 already connected")
                IF (append_file) THEN
                    OPEN(UNIT=9,FILE="NEB_progress.xyz",ACTION='WRITE',POSITION="APPEND",IOSTAT=ios)
                ELSE
                    OPEN(UNIT=9,FILE="NEB_progress.xyz",ACTION='WRITE',IOSTAT=ios)
                    REWIND 9
                ENDIF
                IF (ios/=0) CALL error_handler("can't open output file")
                WRITE(9,*) position_index_2-position_index_1+1
                WRITE(9,*)
                force_constants(:)=force_constant_default
                maxstep=diagonal_length/17.3
                max_displacement=0.0
                skip_position_initialisation=reuse_positions
                IF (.NOT.(ALLOCATED(positions_real))) THEN
                    ALLOCATE(positions_real(0:position_index_2-position_index_1,3),STAT=allocstatus)
                    IF (allocstatus/=0) CALL error_handler("cannot allocate positions_real memory")
                    IF (reuse_positions) THEN
                        WRITE(*,*) "cannot re-use previous positions - none available yet."
                        WRITE(*,*) "Will keep positions for next round."
                        skip_position_initialisation=.FALSE.
                    ENDIF
                ENDIF
                DO n=0,position_index_2-position_index_1,1
                    IF (.NOT.(skip_position_initialisation)) positions_real(n,:)=get_vector(position_index_1+n)
                    WRITE(9,*) "X ",positions_real(n,:)
                ENDDO
            END SUBROUTINE initialise_NEB

            SUBROUTINE calculate_tangents()
            IMPLICIT NONE
            INTEGER :: n
            REAL :: V,Vprev,Vnext,DVmax,DVmin,tplus(3),tminus(3)
                DO n=1,position_index_2-position_index_1-1,1
                    Vprev=givevalue_index_averaged(position_index_1+n-1)
                    V=givevalue_index_averaged(position_index_1+n)
                    Vnext=givevalue_index_averaged(position_index_1+n+1)
                    DVmax=MAX(ABS(Vnext-V),ABS(Vprev-V))
                    DVmin=MIN(ABS(Vnext-V),ABS(Vprev-V))
                    tplus=positions_real(n+1,:)-positions_real(n,:)
                    tminus=positions_real(n,:)-positions_real(n-1,:)
                    IF ((Vnext>V).AND.(V>Vprev)) THEN
                        !Vi+1>V>Vi-1 --> use t+
                        tangents(n,:)=tplus
                    ELSEIF ((Vnext<V).AND.(V<Vprev))THEN
                        !Vi+1<V<Vi-1 --> use t-
                        tangents(n,:)=tminus
                    ELSEIF (Vnext>Vprev) THEN
                        tangents(n,:)=tplus*DVmax+tminus*DVmin
                    ELSEIF (Vnext<=Vprev) THEN
                        tangents(n,:)=tplus*DVmin+tminus*DVmax
                    ELSE
                        CALL error_handler("exception in calculate_tangents")
                    ENDIF
                    CALL normalize3D(tangents(n,:))
                ENDDO
            END SUBROUTINE calculate_tangents

            SUBROUTINE calculate_forces()
            IMPLICIT NONE
            INTEGER :: n
            REAL :: clipboard,vector(3)
                DO n=1,position_index_2-position_index_1-1,1
                    vector=positions_real(n+1,:)-positions_real(n,:)
                    clipboard=SQRT(SUM(vector(:)**2))
                    vector=positions_real(n,:)-positions_real(n-1,:)
                    clipboard=clipboard-SQRT(SUM(vector(:)**2))
                    Fparallel(n,:)=force_constants(n)*clipboard*tangents(n,:)
                    !project out the parallel part
                    vector=give_gradient(position_index_1+n)
                    vector=vector-(DOT_PRODUCT(vector,tangents(n,:)))*tangents(n,:)
                    Fperpendicular(n,:)=-vector
                    IF (dynamic_force_constants) THEN
                        IF (SQRT(SUM(Fparallel(n,:)**2))<0.001) THEN
                            ForceRatio(n)=1.0
                        ELSE
                            ForceRatio(n)=SQRT(SUM(Fperpendicular(n,:)**2))/SQRT(SUM(Fparallel(n,:)**2))
                        ENDIF
                    ENDIF
                ENDDO
                IF (dynamic_force_constants) THEN
                    average_force_constants=average_force_constants+SUM(force_constants)/SIZE(force_constants)
                    !WRITE(*,FMT='("updating force constants -")',ADVANCE="NO")
                    force_constants(:)=force_constants(:)*SUM(ForceRatio)/SIZE(ForceRatio)
                    !WRITE(*,FMT='("complete.")')
                ENDIF
            END SUBROUTINE calculate_forces

            FUNCTION give_gradient(position_index)
            IMPLICIT NONE
            REAL :: give_gradient(3)
            INTEGER,INTENT(IN) :: position_index
            INTEGER :: position_shifted(3),n
                DO n=1,3,1
                    position_shifted=positions(position_index,:)
                    position_shifted(n)=position_shifted(n)+1
                    give_gradient(n)=givevalue_positions_averaged(position_shifted)
                    position_shifted(n)=position_shifted(n)-2
                    give_gradient(n)=give_gradient(n)-givevalue_positions_averaged(position_shifted)
                ENDDO
                give_gradient(:)=give_gradient(:)/gridsize(:)*2.0
            END FUNCTION give_gradient

            SUBROUTINE displace()
            IMPLICIT NONE
            INTEGER :: n
            REAL :: vector(3),displacement(3),clipboard,randomdisplacement(3)
                DO n=1,position_index_2-position_index_1-1,1
                    vector=positions_real(n,:)
                    displacement=Fparallel(n,:)+Fperpendicular(n,:)
                    clipboard=SQRT(SUM(displacement(:)**2))
                    IF (clipboard>maxstep) THEN
                        CALL normalize3D(displacement)
                        displacement(:)=displacement(:)*maxstep
                    ENDIF
                    randomdisplacement(1)=RAND()-0.5
                    randomdisplacement(2)=RAND()-0.5
                    randomdisplacement(3)=RAND()-0.5
                    CALL normalize3D(randomdisplacement)
                    IF (.NOT.((decreasing_forceconstants).OR.(decreasing_stepsize)))&
                    &displacement(:)=displacement(:)+randomdisplacement*maxstep/100.0
                    IF (clipboard>max_displacement) max_displacement=clipboard
                    positions_real(n,:)=vector+displacement
                    CALL set_vector(positions_real(n,:),position_index_1+n)
                ENDDO
            END SUBROUTINE displace

            SUBROUTINE print_step()
            IMPLICIT NONE
            INTEGER :: n
                WRITE(9,*) position_index_2-position_index_1+1
                WRITE(9,*) "step ",iterations
                DO n=0,position_index_2-position_index_1,1
                    WRITE(9,*) "X ",positions_real(n,:)
                ENDDO
            END SUBROUTINE print_step

            RECURSIVE SUBROUTINE print_recursive(position1,position2)
            IMPLICIT NONE
            REAL,INTENT(IN) :: position1(3),position2(3)
            REAL :: middle(3)
                IF (ALL((INT((position1(:)-origin(:))/gridsize(:)))==(INT((position2(:)-origin(:))/gridsize(:))))) THEN
                    RETURN
                ENDIF
                IF (SQRT(SUM((position1-position2)**2))<(diagonal_length/4.0)) RETURN
                middle=(position1+position2)/2.0
                CALL print_recursive(position1,middle)
                IF (ALL((INT((position1(:)-origin(:))/gridsize(:)))/=(INT((middle(:)-origin(:))/gridsize(:))))) THEN
                    IF (ALL((INT((middle(:)-origin(:))/gridsize(:)))/=(INT((position2(:)-origin(:))/gridsize(:))))) THEN
                        !both are different to the middle one - it must be new!
                        CALL set_vector(middle,dummy)
                        WRITE(11,*) "R ",givevalue_index_averaged(dummy),&
                        &givevalue_index(dummy),-LOG(ABS(givevalue_index(dummy)))*8.314E-3*Temperature
                    ENDIF
                ENDIF
                CALL print_recursive(middle,position2)
            END SUBROUTINE

            SUBROUTINE print_values()
            IMPLICIT NONE
            LOGICAL :: connected
            INTEGER :: ios,n
                INQUIRE(UNIT=11,OPENED=connected)
                IF (connected) CALL error_handler("unit 11 already connected")
                OPEN(UNIT=11,FILE="PMF_fine.dat",ACTION='WRITE',IOSTAT=ios)
                IF (ios/=0) CALL error_handler("can't open PMF file")
                WRITE(11,*) "Point SDF_value_averaged SDF_value PMF/kJmol-1"
                DO n=0,position_index_2-position_index_1-1,1
                    CALL set_vector(positions_real(n,:),dummy)
                    WRITE(11,*) n,givevalue_index_averaged(dummy),&
                    &givevalue_index(dummy),-LOG(ABS(givevalue_index(dummy)))*8.314E-3*Temperature
                    CALL print_recursive(positions_real(n,:),positions_real(n+1,:))
                ENDDO
                CALL set_vector(positions_real(position_index_2-position_index_1,:),dummy)
                WRITE(11,*) n,givevalue_index_averaged(dummy),&
                &givevalue_index(dummy),-LOG(ABS(givevalue_index(dummy)))*8.314E-3*Temperature
                CLOSE(UNIT=11)
            END SUBROUTINE print_values

        END SUBROUTINE NEB_segment

        SUBROUTINE normalize3D(vector)!normalizes a 3D vector
        IMPLICIT NONE
        REAL,INTENT(INOUT) :: vector(3)
        REAL :: length
            length=SQRT((vector(1))**2+(vector(2))**2+(vector(3))**2)
            IF (length==0.0) THEN
                vector(:)=0.0
                !PRINT *,"SET TO ZERO"
            ELSE
                vector=(vector/length)
            ENDIF
        END SUBROUTINE normalize3D

        SUBROUTINE print_coordinates(position_index_1,position_index_2)
        IMPLICIT NONE
        LOGICAL :: connected
        INTEGER :: ios,n
        INTEGER,INTENT(IN) :: position_index_1,position_index_2
            INQUIRE(UNIT=11,OPENED=connected)
            IF (connected) CALL error_handler("unit 11 already connected")
            OPEN(UNIT=11,FILE="coordinates.xyz",ACTION='WRITE',IOSTAT=ios)
            IF (ios/=0) CALL error_handler("can't open coordinates file")
            WRITE(11,*) position_index_2-position_index_1+1
            WRITE(11,*)
            DO n=position_index_1,position_index_2,1
                WRITE(11,*) "X ",get_vector(n)
            ENDDO
            CLOSE(UNIT=11)
        END SUBROUTINE print_coordinates

        SUBROUTINE print_NEBSTRING_coordinates()
        IMPLICIT NONE
            CALL print_coordinates(1,NEB_string(SIZE(NEB_string(:))))
        END SUBROUTINE print_NEBSTRING_coordinates

        SUBROUTINE print_values_global(position_index_1,position_index_2)
        IMPLICIT NONE
        LOGICAL :: connected
        INTEGER :: ios,n
        INTEGER,INTENT(IN) :: position_index_1,position_index_2
            INQUIRE(UNIT=11,OPENED=connected)
            IF (connected) CALL error_handler("unit 11 already connected")
            OPEN(UNIT=11,FILE="PMF.dat",ACTION='WRITE',IOSTAT=ios)
            IF (ios/=0) CALL error_handler("can't open PMF file")
            WRITE(11,*) "Point SDF_value_averaged SDF_value PMF/kJmol-1"
            DO n=position_index_1,position_index_2,1
                IF (givevalue_index(n)<0.0) THEN
                    WRITE(11,*) n,givevalue_index_averaged(n),givevalue_index(n),-LOG(-givevalue_index(n))*8.314E-3*Temperature
                ELSE
                    WRITE(11,*) n,givevalue_index_averaged(n),givevalue_index(n),-LOG(givevalue_index(n))*8.314E-3*Temperature
                ENDIF
            ENDDO
            CLOSE(UNIT=11)
        END SUBROUTINE print_values_global

        SUBROUTINE read_vectors(filename,maximise_endpoints)
        IMPLICIT NONE
        CHARACTER(LEN=*),INTENT(IN) :: filename
        LOGICAL,INTENT(IN) :: maximise_endpoints
        INTEGER :: Nvectors,ios,n,allocstatus
        LOGICAL :: connected
            WRITE(*,'("Initialising vectors from file: ",A)') TRIM(filename) 
            INQUIRE(UNIT=9,OPENED=connected)
            IF (connected) CALL error_handler("unit 9 already connected")
            OPEN(UNIT=9,FILE=TRIM(filename) ,ACTION='READ',IOSTAT=ios)
            IF (ios/=0) CALL error_handler("can't open vectors file")
            READ(9,IOSTAT=ios,FMT=*) Nvectors
            IF (ios/=0) CALL error_handler("cannot read header of vectors file")
            READ(9,IOSTAT=ios,FMT=*)
            IF (ios/=0) CALL error_handler("cannot read header of vectors file")
            ALLOCATE(NEB_string(Nvectors),STAT=allocstatus)
            IF (allocstatus/=0) CALL error_handler("cannot allocate NEB_string memory")
            CALL read_body()
            CLOSE(UNIT=9)

            CONTAINS

                SUBROUTINE read_body()
                IMPLICIT NONE
                REAL :: vectors(Nvectors,3),resolution,distance
                INTEGER :: NBeads,last_position
                CHARACTER(LEN=2) :: dummy
                    resolution=0.5 !Number of beads per diagonal length.
                    DO n=1,Nvectors,1
                        READ(9,IOSTAT=ios,FMT=*) dummy,vectors(n,:)
                        IF (ios/=0) CALL error_handler("cannot read body of vectors file")
                    ENDDO
                    IF (maximise_endpoints) THEN
                        CALL set_vector(vectors(1,:),1)
                        CALL set_vector(vectors(Nvectors,:),2)
                        CALL maximise_vector(1,.TRUE.)
                        CALL maximise_vector(2,.TRUE.)
                        vectors(1,:)=get_vector(1)
                        vectors(Nvectors,:)=get_vector(2)
                    ENDIF
                    last_position=1
                    CALL set_vector(vectors(1,:),last_position)
                    NEB_string(1)=1
                    DO n=2,Nvectors,1
                        distance=SQRT(SUM((vectors(n,:)-vectors(n-1,:))**2))
                        NBeads=CEILING(resolution*distance/diagonal_length)-1
                        IF (NBeads<0) NBeads=0
                        WRITE(*,'("Adding ",I0," beads between vector ",I0," and ",I0,".")') NBeads,n-1,n
                        CALL set_vector(vectors(n,:),last_position+NBeads+1)
                        CALL interpolate_vectors(last_position,last_position+NBeads+1)
                        NEB_string(n)=last_position+NBeads+1
                        last_position=last_position+NBeads+1
                    ENDDO
                    WRITE(*,'("using a total of ",I0," beads.")') last_position
                END SUBROUTINE read_body

        END SUBROUTINE read_vectors

        SUBROUTINE full_NEB(pre_optimise_segments,play_safe)
        IMPLICIT NONE
        INTEGER :: stringlength,number_of_endpoints,n,ios
        LOGICAL,INTENT(IN) :: pre_optimise_segments,play_safe
        LOGICAL :: connected
            INQUIRE(UNIT=9,OPENED=connected)
            IF (connected) CALL error_handler("unit 9 already connected")
            OPEN(UNIT=9,FILE="NEB_progress.xyz",ACTION='WRITE',IOSTAT=ios)
            REWIND 9
            IF (ios/=0) CALL error_handler("can't open output file")
            CLOSE(UNIT=9)
            number_of_endpoints=SIZE(NEB_string(:))
            stringlength=NEB_string(number_of_endpoints)
            CALL invert_cubevalues()
            IF (pre_optimise_segments) THEN
                DO n=1,number_of_endpoints-1,1
                    CALL NEB_segment(NEB_string(n),NEB_string(n+1),.FALSE.,.FALSE.,.TRUE.,20,.FALSE.,.FALSE.)
                ENDDO
            ENDIF
            IF (play_safe) THEN
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.FALSE.,.FALSE.,.FALSE.,400,.FALSE.,.FALSE.)
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.TRUE.,.FALSE.,.TRUE.,100,.FALSE.,.FALSE.)
            ELSE
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.FALSE.,.FALSE.,.FALSE.,5000,.FALSE.,.TRUE.)
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.TRUE.,.FALSE.,.TRUE.,5000,.FALSE.,.TRUE.) !Turn on dynamic force constants
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.FALSE.,.TRUE.,.TRUE.,5000,.TRUE.,.TRUE.) !Turn on decreasing force constants
                CALL NEB_segment(NEB_string(1),NEB_string(number_of_endpoints),.TRUE.,.TRUE.,.TRUE.,5000,.TRUE.,.TRUE.) !Turn on decreasing step size
            ENDIF
            !position_index_1,position_index_2,dynamic_force_constants,decreasing_stepsize,append_file,Niterations,decreasing_forceconstants,reuse_positions)
            CALL print_values_global(1,stringlength)
            CALL invert_cubevalues()
            CALL print_NEBSTRING_coordinates()
        END SUBROUTINE full_NEB

END MODULE cubefile

PROGRAM READCUBE
USE cubefile
LOGICAL :: file_exists
CHARACTER(LEN=1024) :: inputstring
INTEGER :: number_of_basins
    number_of_basins=1
    IF (COMMAND_ARGUMENT_COUNT()<1) THEN
        PRINT *,"Please provide as arguments:"
        PRINT *,"cubefile / number of basins --> basin search"
        PRINT *,"cubefile --> NEB (needs a vectors.xyz file)"
        STOP
    ENDIF
    CALL GET_COMMAND_ARGUMENT(1,inputstring)
    INQUIRE(FILE=TRIM(inputstring),EXIST=file_exists)
    IF (COMMAND_ARGUMENT_COUNT()>1) THEN
        CALL initialise_cubefile(TRIM(inputstring))
		CALL GET_COMMAND_ARGUMENT(2,inputstring)
		READ(inputstring,*) number_of_basins
        CALL fill_maxima(number_of_basins)
        CALL finalise_cubefile()
    ELSE
        CALL initialise_cubefile(TRIM(inputstring))
        CALL read_vectors("vectors.xyz",.TRUE.)
        CALL full_NEB(.FALSE.,.FALSE.)
        CALL finalise_cubefile()
        WRITE(*,*) "USING NEB FROM:"
        WRITE(*,*) "https://aip.scitation.org/doi/pdf/10.1063/1.1323224"
		WRITE(*,'("With Temperature T=",F0.2," K.")') Temperature
    ENDIF
END
